/*-
 * Copyright 2017 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"

#include <stdint.h>
#include <inttypes.h>
#include <string.h>
#include <locale.h>

#include "vcl.h"
#include "cache/cache.h"
#include "vrt.h"
#include "vas.h"
#include "vdef.h"

#include "vcc_if.h"

#define PCRE2_CODE_UNIT_WIDTH 8
#include "pcre2.h"

#define ERR(ctx, msg) \
        errmsg((ctx), "vmod pcre2 error: " msg)

#define VERR(ctx, fmt, ...) \
        errmsg((ctx), "vmod pcre2 error: " fmt, __VA_ARGS__)

#define VERRNOMEM(ctx, fmt, ...) \
        VERR((ctx), fmt ", out of space", __VA_ARGS__)

#define ERRNOMEM(ctx, msg) \
        ERR((ctx), msg ", out of space")

#define INIT(ctx) (((ctx)->method & VCL_MET_INIT) != 0)

#define SET_CTX_PARAM(ctx, param, type)	do {				\
		if ((param) != 0) AZ(pcre2_set_##param((ctx), (type)(param))); \
	} while(0)

#define COMPILE_OPTS							\
	VCL_BOOL allow_empty_class, VCL_BOOL anchored, VCL_ENUM bsrs,	\
		VCL_BOOL alt_bsux, VCL_BOOL alt_circumflex,		\
		VCL_BOOL alt_verbnames, VCL_BOOL caseless,		\
		VCL_BOOL dollar_endonly, VCL_BOOL dotall,		\
		VCL_BOOL dupnames, VCL_BOOL extended, VCL_BOOL firstline, \
		VCL_STRING locale, VCL_BOOL match_unset_backref,	\
		VCL_INT max_pattern_length, VCL_BOOL multiline,		\
		VCL_BOOL never_backslash_c, VCL_BOOL never_ucp,		\
		VCL_BOOL never_utf, VCL_ENUM newlines,			\
		VCL_BOOL no_auto_capture, VCL_BOOL no_auto_possess,	\
		VCL_BOOL no_dotstar_anchor, VCL_BOOL no_start_optimize, \
		VCL_BOOL no_utf_check, VCL_INT parens_nest_limit,	\
		VCL_BOOL ucp, VCL_BOOL ungreedy,			\
		VCL_BOOL use_offset_limit, VCL_BOOL utf

#define COMPILE_CTX_OPTS						\
	VCL_ENUM bsrs, VCL_STRING locale, VCL_INT max_pattern_length,	\
		VCL_ENUM newlines, VCL_INT parens_nest_limit

#define COMPILE_CTX_PARAMS						\
	bsrs, locale, max_pattern_length, newlines, parens_nest_limit

#define COMPILE_FLAGS							\
	VCL_BOOL allow_empty_class, VCL_BOOL anchored, VCL_BOOL alt_bsux, \
		VCL_BOOL alt_circumflex, VCL_BOOL alt_verbnames,	\
		VCL_BOOL caseless, VCL_BOOL dollar_endonly,		\
		VCL_BOOL dotall, VCL_BOOL dupnames, VCL_BOOL extended,	\
		VCL_BOOL firstline, VCL_BOOL match_unset_backref,	\
		VCL_BOOL multiline, VCL_BOOL never_backslash_c,		\
		VCL_BOOL never_ucp, VCL_BOOL never_utf,			\
		VCL_BOOL no_auto_capture, VCL_BOOL no_auto_possess,	\
		VCL_BOOL no_dotstar_anchor, VCL_BOOL no_start_optimize, \
		VCL_BOOL no_utf_check, VCL_BOOL ucp, VCL_BOOL ungreedy,	\
		VCL_BOOL use_offset_limit, VCL_BOOL utf

#define COMPILE_FLAGS_PARAMS						\
	allow_empty_class, anchored, alt_bsux, alt_circumflex,		\
		alt_verbnames, caseless, dollar_endonly, dotall,	\
		dupnames, extended, firstline, match_unset_backref,	\
		multiline, never_backslash_c, never_ucp, never_utf,	\
		no_auto_capture, no_auto_possess, no_dotstar_anchor,	\
		no_start_optimize, no_utf_check, ucp, ungreedy,		\
		use_offset_limit, utf

#define MATCH_OPTS							\
	VCL_INT len, VCL_BOOL anchored, VCL_INT match_limit,		\
		VCL_INT offset_limit, VCL_BOOL notbol, VCL_BOOL noteol, \
		VCL_BOOL notempty, VCL_BOOL notempty_atstart,		\
		VCL_BOOL no_jit, VCL_BOOL no_utf_check, VCL_ENUM partials, \
		VCL_INT recursion_limit

/* Doesn't repeat the anchored and no_utf_check options */
#define MATCHF_OPTS							\
	VCL_INT len, VCL_INT match_limit, VCL_INT offset_limit,		\
		VCL_BOOL notbol, VCL_BOOL noteol, VCL_BOOL notempty,	\
		VCL_BOOL notempty_atstart, VCL_BOOL no_jit,		\
		VCL_ENUM partials, VCL_INT recursion_limit

#define MATCH_CTX_OPTS \
	VCL_INT match_limit, VCL_INT offset_limit, VCL_INT recursion_limit

#define MATCH_CTX_PARAMS \
	match_limit, offset_limit, recursion_limit

#define MATCH_FLAGS							\
	VCL_BOOL anchored, VCL_BOOL notbol, VCL_BOOL noteol,		\
		VCL_BOOL notempty, VCL_BOOL notempty_atstart,		\
		VCL_BOOL no_jit, VCL_BOOL no_utf_check, VCL_ENUM partials

#define MATCH_FLAGS_PARAMS						\
	anchored, notbol, noteol, notempty, notempty_atstart, no_jit,	\
		no_utf_check, partials

struct vmod_pcre2_regex {
	unsigned	magic;
#define VMOD_PCRE2_REGEX_MAGIC 0x3adb2a78
	pcre2_code	*code;
	char		*vcl_name;
};

/*
 * PRIV_TASK scope. The general context is used by all methods and
 * functions.  The match_data is used by the match(), backref() and
 * namedref() functions.
 */
struct task {
	unsigned		magic;
#define VMOD_PCRE2_TASK_MAGIC 0x21000676
	pcre2_general_context	*gctx;
	pcre2_match_data	*mdata;
};

/*
 * PRIV_CALL scope. The match context and options are used by the match
 * method and function. The compile context and options are used by the
 * match function.
 */
struct match_call {
	unsigned		magic;
#define VMOD_PCRE2_MATCH_CALL_MAGIC 0x60e5bd33
	pcre2_match_context	*mctx;
	pcre2_compile_context	*cctx;
	uint32_t		match_options;
	uint32_t		compile_options;
};

enum ref_e {
	NUMBERED = 0,
	NAMED = 1,
};

static int have_jit = 0;

void
errmsg(VRT_CTX, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	if (ctx->vsl)
		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
	else
		VSLv(SLT_VCL_Error, 0, fmt, args);
	va_end(args);

	if (ctx->method == VCL_MET_INIT) {
		AN(ctx->msg);
		va_start(args, fmt);
		VSB_vprintf(ctx->msg, fmt, args);
		VSB_putc(ctx->msg, '\n');
		va_end(args);
		VRT_handling(ctx, VCL_RET_FAIL);
	}
}

static void
report_pcre2_err(VRT_CTX, int errcode, const char * const restrict msg,
		 const char * const restrict post)
{
	int ret;

	uintptr_t snap = WS_Snapshot(ctx->ws);
	unsigned bytes = WS_Reserve(ctx->ws, 0);
	char *buf = WS_Front(ctx->ws);

	ret = pcre2_get_error_message(errcode, (PCRE2_UCHAR *)buf,
				      (PCRE2_SIZE)bytes);
	if (ret == PCRE2_ERROR_BADDATA) {
		WS_ReleaseP(ctx->ws, buf);
		VERR(ctx, "%s (unknown error code)%s", msg, post);
	}
	else if (ret == PCRE2_ERROR_NOMEMORY) {
		unsigned len = strlen(buf) + 1;

		if (len > bytes) {
			WS_Release(ctx->ws, 0);
			ERRNOMEM(ctx, "allocating workspace for pcre2 error "
				 "message");
			VERR(ctx, "%s: libpcre2 error code %d%s", msg, errcode,
			     post);
		}
		else {
			WS_Release(ctx->ws, len);
			VERR(ctx, "%s: %s (truncated error message)%s", msg,
			     buf, post);
		}
	}
	else {
		WS_Release(ctx->ws, ret);
		VERR(ctx, "%s: %s%s", msg, buf, post);
	}
	WS_Reset(ctx->ws, snap);
}

static inline int
check_uint32_range(VRT_CTX, long long limit, const char * const restrict name,
		   const char * const restrict context,
		   const char * const restrict caller)
{
	if (limit < 0 || limit > UINT32_MAX) {
		VERR(ctx, "%s (%lld) out of range in %s%s (must be >= 0 and "
		     "<= %" PRIu32 ")", name, limit, context, caller,
		     UINT32_MAX);
		return 0;
	}
	return 1;
}

/* Event function */

int __match_proto__(vmod_event_f)
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	(void) ctx;
	(void) priv;

	if (e == VCL_EVENT_LOAD) {
		int ret;
		uint32_t where;

		/*
		 * A VCL_INT must never be larger than PCRE2_SIZE_MAX, so
		 * that some assignments from INT params cannot be out of
		 * range.
		 * XXX: C11 static_assert would catch this at compile
		 * time.
		 */
		assert(sizeof(VCL_INT) <= sizeof(PCRE2_SIZE));

		ret = pcre2_config(PCRE2_CONFIG_JIT, &where);
		assert(ret >= 0);
		have_jit = where;
	}
	return 0;
}

/* Object regex */

static void *
ws_malloc(PCRE2_SIZE sz, void *ws)
{
	return WS_Alloc(ws, sz);
}

static void
ws_free(void *ptr, void *ws)
{
	WS_Assert_Allocated(ws, ptr, 0);
}

static void
match_call_free(void *priv)
{
	struct match_call *match_call;

	if (priv == NULL)
		return;
	CAST_OBJ(match_call, priv, VMOD_PCRE2_MATCH_CALL_MAGIC);
	if (match_call->cctx != NULL)
		pcre2_compile_context_free(match_call->cctx);
	if (match_call->mctx != NULL)
		pcre2_match_context_free(match_call->mctx);
	FREE_OBJ(match_call);
}

static pcre2_compile_context *
get_compile_context(VRT_CTX, COMPILE_CTX_OPTS,
		    const char * restrict const context,
		    const char * restrict const caller)
{
	pcre2_compile_context *ccontext;
	uint32_t val;

	if (!check_uint32_range(ctx, parens_nest_limit, "parens_nest_limit",
				context, caller))
		return NULL;
	if (max_pattern_length < 0) {
		VERR(ctx, "max_pattern_len (%lld) must be >= 0 in %s%s",
		     (long long)max_pattern_length, context, caller);
		return NULL;
	}
	if ((ccontext = pcre2_compile_context_create(NULL)) == NULL) {
		VERR(ctx, "failed to create compile context in %s%s", context,
		     caller);
		return NULL;
	}
	SET_CTX_PARAM(ccontext, max_pattern_length, PCRE2_SIZE);
	SET_CTX_PARAM(ccontext, parens_nest_limit, uint32_t);
	if (bsrs != NULL) {
		if (strcmp("ANYCRLF", bsrs) == 0)
			val = PCRE2_BSR_ANYCRLF;
		else if (strcmp("UNICODE", bsrs) == 0)
			val = PCRE2_BSR_UNICODE;
		else
			WRONG("Illegal bsr enum value");
		AZ(pcre2_set_bsr(ccontext, val));
	}
	if (locale != NULL) {
		const unsigned char *tables;
		char *saved_lc = setlocale(LC_CTYPE, NULL);

		AN(saved_lc);
		if (setlocale(LC_CTYPE, locale) == NULL) {
			AN(setlocale(LC_CTYPE, saved_lc));
			VERR(ctx, "Cannot set locale %s in %s%s", locale,
			     context, caller);
			return NULL;
		}
		tables = pcre2_maketables(NULL);
		if (tables == NULL) {
			AN(setlocale(LC_CTYPE, saved_lc));
			VERRNOMEM(ctx, "Creating tables for locale %s in %s%s",
				  locale, context, caller);
			return NULL;
		}
		AZ(pcre2_set_character_tables(ccontext, tables));
		AN(setlocale(LC_CTYPE, saved_lc));
	}
	if (newlines != NULL) {
		if (strcmp("CR", newlines) == 0)
			val = PCRE2_NEWLINE_CR;
		else if (strcmp("LF", newlines) == 0)
			val = PCRE2_NEWLINE_LF;
		else if (strcmp("CRLF", newlines) == 0)
			val = PCRE2_NEWLINE_CRLF;
		else if (strcmp("ANYCRLF", newlines) == 0)
			val = PCRE2_NEWLINE_ANYCRLF;
		else if (strcmp("ANY", newlines) == 0)
			val = PCRE2_NEWLINE_ANY;
		else
			WRONG("Illegal newline enum value");
		AZ(pcre2_set_newline(ccontext, val));
	}
	return ccontext;
}

static pcre2_match_context *
get_match_context(VRT_CTX, MATCH_CTX_OPTS, const char * restrict const context,
		  const char * restrict const caller)
{
	pcre2_match_context *mctx;

	if (!check_uint32_range(ctx, match_limit, "match_limit", context,
				caller))
		return NULL;
	if (!check_uint32_range(ctx, offset_limit, "offset_limit", context,
				caller))
		return NULL;
	if (!check_uint32_range(ctx, recursion_limit, "recursion_limit",
				context, caller))
		return NULL;

	if ((mctx = pcre2_match_context_create(NULL)) == NULL) {
		VERRNOMEM(ctx, "creating match context in %s%s", context,
			  caller);
		return NULL;
	}
	SET_CTX_PARAM(mctx, match_limit, uint32_t);
	SET_CTX_PARAM(mctx, offset_limit, uint32_t);
	SET_CTX_PARAM(mctx, recursion_limit, uint32_t);
	return mctx;
}

static inline struct task *
get_task(VRT_CTX, struct vmod_priv *priv_task,
	 const char * const restrict context,
	 const char * const restrict caller)
{
	struct task *match_task;

	/*
	 * The general context is task-scoped and shared by all methods
	 * and functions from this VMOD in the task. It uses WS_Alloc as
	 * the custom allocator.
	 */
	if (priv_task->priv == NULL) {
		if ((match_task = WS_Alloc(ctx->ws, sizeof(*match_task)))
		    == NULL) {
			VERRNOMEM(ctx, "allocating task-scoped data in %s%s",
				  context, caller);
			return NULL;
		}
		if ((match_task->gctx
		     = pcre2_general_context_create(ws_malloc, ws_free,
						    ctx->ws)) == NULL) {
			VERRNOMEM(ctx, "initializing general context in %s%s",
				  context, caller);
			return NULL;
		}
		match_task->magic = VMOD_PCRE2_TASK_MAGIC;
		match_task->mdata = NULL;
		priv_task->priv = match_task;
		priv_task->len = sizeof(*match_task);
		priv_task->free = NULL;
	}
	else {
		WS_Assert_Allocated(ctx->ws, priv_task->priv,
				    sizeof(*match_task));
		CAST_OBJ(match_task, priv_task->priv, VMOD_PCRE2_TASK_MAGIC);
		/*
		 * pcre2_general_context is declared as an incomplete
		 * type, so we can't check if its full size is allocated.
		 */
		WS_Assert_Allocated(ctx->ws, match_task->gctx, 0);
	}
	return match_task;
}

static inline void
set_opt(uint32_t *options, VCL_BOOL vmod_opt, uint32_t pcre2_opt)
{
	if (vmod_opt)
		*options |= pcre2_opt;
}

static inline void
set_compile_flags(uint32_t *options, COMPILE_FLAGS)
{
	/* XXX check illegal combinations such as never_ucp && ucp ... ? */
	set_opt(options, anchored,		PCRE2_ANCHORED);
	set_opt(options, allow_empty_class,	PCRE2_ALLOW_EMPTY_CLASS);
	set_opt(options, alt_bsux,		PCRE2_ALT_BSUX);
	set_opt(options, alt_circumflex,	PCRE2_ALT_CIRCUMFLEX);
	set_opt(options, alt_verbnames,		PCRE2_ALT_VERBNAMES);
	set_opt(options, caseless,		PCRE2_CASELESS);
	set_opt(options, dollar_endonly,	PCRE2_DOLLAR_ENDONLY);
	set_opt(options, dotall,		PCRE2_DOTALL);
	set_opt(options, dupnames,		PCRE2_DUPNAMES);
	set_opt(options, extended,		PCRE2_EXTENDED);
	set_opt(options, firstline,		PCRE2_FIRSTLINE);
	set_opt(options, match_unset_backref,	PCRE2_MATCH_UNSET_BACKREF);
	set_opt(options, multiline,		PCRE2_MULTILINE);
	set_opt(options, never_backslash_c,	PCRE2_NEVER_BACKSLASH_C);
	set_opt(options, never_ucp,		PCRE2_NEVER_UCP);
	set_opt(options, never_utf,		PCRE2_NEVER_UTF);
	set_opt(options, no_auto_capture,	PCRE2_NO_AUTO_CAPTURE);
	set_opt(options, no_auto_possess,	PCRE2_NO_AUTO_POSSESS);
	set_opt(options, no_dotstar_anchor,	PCRE2_NO_DOTSTAR_ANCHOR);
	set_opt(options, no_start_optimize,	PCRE2_NO_START_OPTIMIZE);
	set_opt(options, no_utf_check,		PCRE2_NO_UTF_CHECK);
	set_opt(options, ucp,			PCRE2_UCP);
	set_opt(options, ungreedy,		PCRE2_UNGREEDY);
	set_opt(options, use_offset_limit,	PCRE2_USE_OFFSET_LIMIT);
	set_opt(options, utf,			PCRE2_UTF);
}

static inline void
set_match_flags(uint32_t *options, MATCH_FLAGS)
{
	set_opt(options, anchored,		PCRE2_ANCHORED);
	set_opt(options, notbol,		PCRE2_NOTBOL);
	set_opt(options, noteol,		PCRE2_NOTEOL);
	set_opt(options, notempty,		PCRE2_NOTEMPTY);
	set_opt(options, notempty_atstart,	PCRE2_NOTEMPTY_ATSTART);
	set_opt(options, no_jit,		PCRE2_NO_JIT);
	set_opt(options, no_utf_check,		PCRE2_NO_UTF_CHECK);

	/* Partial matching not implemented yet */
	AZ(partials);
}

static pcre2_code *
compile(VRT_CTX, pcre2_compile_context * restrict const cctx,
	VCL_STRING const restrict pattern, uint32_t options, int do_jit,
	const char * const restrict context, const char * const restrict caller)
{
	pcre2_code *code;
	int err_code = 0;
	PCRE2_SIZE err_offset;

	/* XXX set the length via parameter */
	code = pcre2_compile((PCRE2_SPTR)pattern, PCRE2_ZERO_TERMINATED,
			     options, &err_code, &err_offset, cctx);
	if (code == NULL) {
		char *msg, *offset_msg;
		uintptr_t snap = WS_Snapshot(ctx->ws);

		if ((msg = WS_Printf(ctx->ws, "Cannot compile '%s' in %s%s",
				     pattern, context, caller)) == NULL)
			msg = "";
		if ((offset_msg = WS_Printf(ctx->ws, " at offset %zu",
					    err_offset)) == NULL)
			offset_msg = "";
		report_pcre2_err(ctx, err_code, msg, offset_msg);
		WS_Reset(ctx->ws, snap);
		return NULL;
	}
	if (do_jit) {
		int ret;

		/* XXX check option compatibility; disable via param */
		/* XXX set complete or soft/hard partial via param */
		options |= PCRE2_JIT_COMPLETE;
		ret = pcre2_jit_compile(code, options);
		if (ret != 0) {
			char *msg;
			uintptr_t snap = WS_Snapshot(ctx->ws);

			if ((msg = WS_Printf(ctx->ws, "Cannot jit-compile "
					     "'%s' in %s%s", pattern, context,
					     caller)) == NULL)
				msg = "";
			report_pcre2_err(ctx, ret, msg, "");
			WS_Reset(ctx->ws, snap);
			return NULL;
		}
	}
	return code;
}

static inline VCL_BOOL
match(VRT_CTX, pcre2_code *code, VCL_STRING subject, VCL_INT len,
      uint32_t options, pcre2_match_data *mdata, pcre2_match_context *mctx,
      const char * restrict const context, const char * restrict const caller)
{
	int ret;
	uintptr_t snap;
	char *msg;

	if (subject == NULL)
		subject = "";
	if (len == 0)
		len = PCRE2_ZERO_TERMINATED;

	/* XXX param for start_offset */
	ret = pcre2_match(code, (PCRE2_SPTR)subject, (PCRE2_SIZE)len, 0,
			  options, mdata, mctx);
	if (ret == PCRE2_ERROR_NOMATCH)
		return 0;
	assert(ret != PCRE2_ERROR_PARTIAL); /* XXX */
	if (ret >= 0)
		return 1;

	snap = WS_Snapshot(ctx->ws);
	if ((msg = WS_Printf(ctx->ws, "in %s%s", context, caller)) == NULL)
		msg = "";
	report_pcre2_err(ctx, ret, msg, "");
	WS_Reset(ctx->ws, snap);
	return 0;
}

static inline VCL_STRING
refer(VRT_CTX, VCL_INT ref, VCL_STRING restrict const name,
      VCL_STRING restrict const fallback, const enum ref_e reftype,
      pcre2_match_data *mdata, const char * restrict const context,
      const char * restrict const caller,
      const char * restrict const default_fallback)
{

	PCRE2_UCHAR *buf;
	PCRE2_SIZE len;
	int ret = PCRE2_ERROR_UNAVAILABLE;
	char *msg = NULL;
	uintptr_t snap;

	if (fallback == NULL) {
		VERR(ctx, "in %s.%s(): fallback is undefined", context, caller);
		return default_fallback;
	}
	if (mdata == NULL) {
		VERR(ctx, "%s.%s() called without prior match", context,
		     caller);
		return fallback;
	}
	WS_Assert_Allocated(ctx->ws, mdata, 0);

	switch (reftype) {
	case NUMBERED:
		if (!check_uint32_range(ctx, ref, "ref", context, ".backref()"))
			return fallback;
		ret = pcre2_substring_get_bynumber(mdata, (uint32_t)ref, &buf,
						   &len);
		break;
	case NAMED:
		if (name == NULL) {
			VERR(ctx, "in %s.%s(<undefined>): name is undefined",
			     context, caller);
			return fallback;
		}
		ret = pcre2_substring_get_byname(mdata, (PCRE2_SPTR)name, &buf,
						 &len);
	}

	if (ret == 0) {
		WS_Assert_Allocated(ctx->ws, buf, len + 1);
		return (VCL_STRING)buf;
	}

	/*
	 * This error is returned when the ovector was too small, cannot
	 * happen after using pcre2_match_data_create_from_pattern() to
	 * get the match data block.
	 */
	assert(ret != PCRE2_ERROR_UNAVAILABLE);

	snap = WS_Snapshot(ctx->ws);
	if ((msg = WS_Printf(ctx->ws, "in %s.%s()", context, caller)) == NULL)
		msg = "";
	report_pcre2_err(ctx, ret, msg, "");
	WS_Reset(ctx->ws, snap);
	return fallback;
}

VCL_VOID
vmod_regex__init(VRT_CTX, struct vmod_pcre2_regex **regexp,
		 const char *vcl_name, VCL_STRING pattern, COMPILE_OPTS)
{
	struct vmod_pcre2_regex *regex;
	pcre2_code *code;
	pcre2_compile_context *ccontext;
	uint32_t options = 0;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(regexp);
	AZ(*regexp);
	AN(vcl_name);
	if (pattern == NULL) {
		VERR(ctx, "pattern is NULL in %s constructor", vcl_name);
		return;
	}
	if ((ccontext = get_compile_context(ctx, COMPILE_CTX_PARAMS, vcl_name,
					    " constructor")) == NULL)
		return;
	set_compile_flags(&options, COMPILE_FLAGS_PARAMS);

	if ((code = compile(ctx, ccontext, pattern, options, have_jit,
			    vcl_name, " constructor")) == NULL)
	    return;

	ALLOC_OBJ(regex, VMOD_PCRE2_REGEX_MAGIC);
	AN(regex);
	*regexp = regex;
	regex->code = code;
	regex->vcl_name = strdup(vcl_name);
	AN(regex->vcl_name);
	pcre2_compile_context_free(ccontext);
}

VCL_VOID
vmod_regex__fini(struct vmod_pcre2_regex **regexp)
{
	struct vmod_pcre2_regex *regex;

	if (regexp == NULL || *regexp == NULL)
		return;
	regex = *regexp;
	CHECK_OBJ(regex, VMOD_PCRE2_REGEX_MAGIC);
	if (regex->vcl_name != NULL)
		free(regex->vcl_name);
	pcre2_code_free(regex->code);
	FREE_OBJ(regex);
}

VCL_BOOL
vmod_regex_match(VRT_CTX, struct vmod_pcre2_regex *regex,
		 struct vmod_priv *priv_call, struct vmod_priv *priv_task,
		 VCL_STRING subject, MATCH_OPTS)
{
	pcre2_match_data *mdata;
	struct task *match_task = NULL;
	struct vmod_priv *obj_task = NULL;
	struct match_call *match_opts;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(regex, VMOD_PCRE2_REGEX_MAGIC);
	AN(priv_task);
	AN(priv_call);

	if ((match_task = get_task(ctx, priv_task, regex->vcl_name, ".match()"))
	    == NULL)
		return 0;

	/* The match context and options are PRIV_CALL-scoped. */
	/* XXX param to decide whether these should be saved for the call */
	if (priv_call->priv == NULL) {
		pcre2_match_context *mctx;

		if ((mctx = get_match_context(ctx, MATCH_CTX_PARAMS,
					      regex->vcl_name, ".match()"))
		    == NULL)
			return 0;

		ALLOC_OBJ(match_opts, VMOD_PCRE2_MATCH_CALL_MAGIC);
		AN(match_opts);
		match_opts->mctx = mctx;
		priv_call->priv = match_opts;
		priv_call->free = match_call_free;

		set_match_flags(&match_opts->match_options, MATCH_FLAGS_PARAMS);
	}
	else
		CAST_OBJ(match_opts, priv_call->priv,
			 VMOD_PCRE2_MATCH_CALL_MAGIC);

	/* The match data block is task-scoped for this object only. */
	obj_task = VRT_priv_task(ctx, regex);
	AN(obj_task);
	if (obj_task->priv == NULL) {
		mdata = pcre2_match_data_create_from_pattern(regex->code,
							     match_task->gctx);
		if (mdata == NULL) {
			VERRNOMEM(ctx, "initializing match data in "
				  "%s.match()", regex->vcl_name);
			return 0;
		}
		obj_task->priv = mdata;
		obj_task->free = NULL;
	}
	else {
		/*
		 * As with the general context above, pcre2_match_data is
		 * an incomplete type.
		 */
		WS_Assert_Allocated(ctx->ws, obj_task->priv, 0);
		mdata = obj_task->priv;
	}

	return match(ctx, regex->code, subject, len, match_opts->match_options,
		     mdata, match_opts->mctx, regex->vcl_name, ".match()");
}

VCL_STRING
vmod_regex_backref(VRT_CTX, struct vmod_pcre2_regex *regex, VCL_INT ref,
		   VCL_STRING fallback)
{
	struct vmod_priv *obj_task;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(regex, VMOD_PCRE2_REGEX_MAGIC);

	obj_task = VRT_priv_task(ctx, regex);
	AN(obj_task);
	return refer(ctx, ref, NULL, fallback, NUMBERED,
		     obj_task->priv, regex->vcl_name, "backref",
		     "**BACKREF METHOD FAILED**");
}

VCL_STRING
vmod_regex_namedref(VRT_CTX, struct vmod_pcre2_regex *regex, VCL_STRING name,
		    VCL_STRING fallback)
{
	struct vmod_priv *obj_task;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(regex, VMOD_PCRE2_REGEX_MAGIC);

	obj_task = VRT_priv_task(ctx, regex);
	AN(obj_task);
	return refer(ctx, -1, name, fallback, NAMED, obj_task->priv,
		     regex->vcl_name, "namedref", "**NAMEDREF METHOD FAILED**");
}

/* Functional interface */

VCL_BOOL
vmod_match(VRT_CTX, struct vmod_priv *call, struct vmod_priv *task,
	   VCL_STRING pattern, VCL_STRING subject, COMPILE_OPTS, MATCHF_OPTS)
{
	struct match_call *match_call;
	pcre2_code *code;
	struct task *match_task;
	VCL_BOOL ret = 0;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(call);
	AN(task);
	if (pattern == NULL) {
		ERR(ctx, "pattern is NULL in pcre2.match()");
		return 0;
	}

	/* Compile and match options saved in PRIV_CALL scope */
	/* XXX need a lock */
	if (call->priv == NULL) {
		pcre2_compile_context *cctx;
		pcre2_match_context *mctx;

		if ((cctx = get_compile_context(ctx, COMPILE_CTX_PARAMS,
						"pcre2", ".match()")) == NULL)
			return 0;
		if ((mctx = get_match_context(ctx, MATCH_CTX_PARAMS,
					      "pcre2", ".match()")) == NULL)
			return 0;
		ALLOC_OBJ(match_call, VMOD_PCRE2_MATCH_CALL_MAGIC);
		if (match_call == NULL) {
			ERRNOMEM(ctx, "allocating call-scoped options in "
				 "regex.match()");
			return 0;
		}
		call->priv = match_call;
		call->free = match_call_free;
		match_call->cctx = cctx;
		match_call->mctx = mctx;
		set_compile_flags(&match_call->compile_options,
				  COMPILE_FLAGS_PARAMS);
		set_match_flags(&match_call->match_options, MATCH_FLAGS_PARAMS);
	}
	else
		CAST_OBJ(match_call, call->priv, VMOD_PCRE2_MATCH_CALL_MAGIC);

	if ((code = compile(ctx, match_call->cctx, pattern,
			    match_call->compile_options, have_jit && !no_jit,
			    "pcre2", ".match()")) == NULL)
		return 0;

	if ((match_task = get_task(ctx, task, "pcre2", ".match()")) == NULL)
		return 0;

	match_task->mdata
		= pcre2_match_data_create_from_pattern(code, match_task->gctx);
	if (match_task->mdata == NULL) {
		ERRNOMEM(ctx, "initializing match data in pcre2.match()");
		return 0;
	}
	WS_Assert_Allocated(ctx->ws, match_task->mdata, 0);

	ret = match(ctx, code, subject, len, match_call->match_options,
		    match_task->mdata, match_call->mctx, "pcre2", ".match()");

	pcre2_code_free(code);
	return ret;
}

/* Functions */

VCL_BOOL
vmod_config_bool(VRT_CTX, VCL_ENUM configs)
{
	uint32_t what, where;
	int ret;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	
	if (strcmp(configs, "JIT") == 0)
		what = PCRE2_CONFIG_JIT;
	else if (strcmp(configs, "STACKRECURSE") == 0)
		what = PCRE2_CONFIG_STACKRECURSE;
	else if (strcmp(configs, "UNICODE") == 0)
		what = PCRE2_CONFIG_UNICODE;
	else
		WRONG("Illegal config enum in config_bool");

	ret = pcre2_config(what, &where);
	assert(ret >= 0);
	return (VCL_BOOL) where;
}

VCL_STRING
vmod_config_str(VRT_CTX, VCL_ENUM configs)
{
	uint32_t what;
	int ret;
	void *buf;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (strcmp(configs, "BSR") == 0)
		what = PCRE2_CONFIG_BSR;
	else if (strcmp(configs, "JITTARGET") == 0)
		what = PCRE2_CONFIG_JITTARGET;
	else if (strcmp(configs, "NEWLINE") == 0)
		what = PCRE2_CONFIG_NEWLINE;
	else if (strcmp(configs, "UNICODE_VERSION") == 0)
		what = PCRE2_CONFIG_UNICODE_VERSION;
	else if (strcmp(configs, "VERSION") == 0)
		what = PCRE2_CONFIG_VERSION;
	else
		WRONG("Illegal config enum in config_str");


	if (what == PCRE2_CONFIG_NEWLINE || what == PCRE2_CONFIG_BSR) {
		uint32_t where;

		ret = pcre2_config(what, &where);
		assert(ret >= 0);

		if (what == PCRE2_CONFIG_NEWLINE)
			switch(where) {
			case PCRE2_NEWLINE_CR:
				return "CR";
			case PCRE2_NEWLINE_LF:
				return "LF";
			case PCRE2_NEWLINE_CRLF:
				return "CRLF";
			case PCRE2_NEWLINE_ANY:
				return "ANY";
			case PCRE2_NEWLINE_ANYCRLF:
				return "ANYCRLF";
			default:
				WRONG("Illegal result for "
				      "PCRE2_CONFIG_NEWLINE");
			}

		assert (what == PCRE2_CONFIG_BSR);
		switch(where) {
		case PCRE2_BSR_UNICODE:
			return "UNICODE";
		case PCRE2_BSR_ANYCRLF:
			return "ANYCRLF";
		default:
			WRONG("Illegal result for PCRE2_CONFIG_BSR");
		}
	}

	ret = pcre2_config(what, NULL);
	if (what == PCRE2_CONFIG_JITTARGET && ret == PCRE2_ERROR_BADOPTION)
		return "JIT not supported";
	assert(ret > 0);
	if ((buf = WS_Alloc(ctx->ws, ret)) == NULL) {
		ERRNOMEM(ctx, "allocating result in pcre2.config_str()");
		return NULL;
	}
	ret = pcre2_config(what, buf);
	assert(ret > 0);
	return (VCL_STRING) buf;
}

VCL_INT
vmod_config_int(VRT_CTX, VCL_ENUM configs)
{
	uint32_t what, where;
	int ret;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (strcmp(configs, "LINKSIZE") == 0)
		what = PCRE2_CONFIG_LINKSIZE;
	else if (strcmp(configs, "MATCHLIMIT") == 0)
		what = PCRE2_CONFIG_MATCHLIMIT;
	else if (strcmp(configs, "PARENSLIMIT") == 0)
		what = PCRE2_CONFIG_PARENSLIMIT;
	else if (strcmp(configs, "RECURSIONLIMIT") == 0)
		what = PCRE2_CONFIG_RECURSIONLIMIT;
	else
		WRONG("Illegal config enum in config_int");

	ret = pcre2_config(what, &where);
	assert(ret >= 0);
	return (VCL_INT) where;
}

VCL_STRING
vmod_version(VRT_CTX __attribute__((unused)))
{
	return VERSION;
}
